#include <stdio.h>
#include <stdlib.h>
#include "types.h"
#include "polygon.h"
#include "vl.h"

/* Matrix multiplies Xform by SourceVec and stores the result in */
/* DestVec. Multiplies a 4 x 4 matrix times a 4 x 1 matrix; the  */
/* result is a 4 x 1 matrix, as follows:                         */
/* --   --    --  --   --  --                                    */
/* |     |     | 4 |    | 4 |                                    */ 
/* | 4x4 |  x  | x | =  | x |                                    */
/* |     |     | 1 |    | 1 |                                    */
/* --   --    --  --   --  --                                    */
void XformVec(double Xform[4][4], double *SourceVec, double *DestVec)
{
    S32 i;
    S32 j;
    
    for(i = 0; i < 4; i++)
    {
        DestVec[i] = 0;  
        for(j = 0; j < 4; j++)
        {
            DestVec[i] += Xform[i][j] * SourceVec[j];
        }     
    }     
}

/* Matrix multiplies SourceXform1 by SourceXform2 and stores the */
/* result in DestXform. Multiplies a 4 x 4 matrix times a 4 x 4  */
/* matrix: the result is a 4 x 4 matrix, as follows:             */
/* --   --     --     --   --  --                                */
/* |     |     |       |    | 4 |                                */ 
/* | 4x4 |  x  | 4 x 4 | =  | x |                                */
/* |     |     |       |    | 1 |                                */
/* --   --     --     --   --  --                                */
void ConcatXforms(const double SourceXform1[4][4], const double SourceXform2[4][4],double (*DestXform)[4])
{
    S32 i;
    S32 j;
    S32 k;
    
    for(i = 0; i < 4; i++)
    {
        for(j=0 ; j < 4; j++)
        {
            /* clear dest before starting matrix concat*/
            DestXform[i][j] = 0;
            /* concat matrixs */
            for(k = 0; k < 4; k++)
            {
                DestXform[i][j] += SourceXform1[i][k] * SourceXform2[k][j];
            }
        }
    } 
}

/* Transforms all vertices in the specified object into view space, then    */
/* perspective projects them to screen space and maps them to screen screen */
/* coordinates storing the result in the object                             */
void XformAndProjectPoints(double Xform[4][4], struct Object *ObjectToXform)
{
    S32 i;
    S32 NumPoints = ObjectToXform->NumVerts;
    struct Point3 *Points = ObjectToXform->VertexList; 
    struct Point3 *XformedPoints = ObjectToXform->XformedVertexList;
    struct Point3 *ProjectedPoints = ObjectToXform->ProjectedVertexList;
    struct Point  *ScreenPoints = ObjectToXform->ScreenVertexList;
    
    /* transforms to view space, then project to the screen */
    for(i = 0; i < NumPoints; i++,Points++,XformedPoints++,ProjectedPoints++,ScreenPoints++)
    {
        /* transform to view space */
        XformVec(Xform, (double *)Points, (double *)XformedPoints);
        /* perspective-project to screen space */
        ProjectedPoints->X = XformedPoints->X / XformedPoints->Z * PROJECTION_RATIO * (SCREEN_WIDTH / 2.0);
        ProjectedPoints->Y = XformedPoints->Y / XformedPoints->Z * PROJECTION_RATIO * (SCREEN_WIDTH / 2.0);
        ProjectedPoints->Z = XformedPoints->Z;
        /* Convert to screen coorinates. The Y coord is negated to flip from */
        /* increasing Y being up to increasing Y being down, as expected by  */
        /* the polygon filler. Add in half the screen width and height to    */
        /* centre on the screen                                              */
        ScreenPoints->X = ((S32) floor(ProjectedPoints->X + 0.5)) + SCREEN_WIDTH / 2; 
        ScreenPoints->Y = (-((S32) floor(ProjectedPoints->Y + 0.5))) + SCREEN_HEIGHT / 2;                    
    }
}

/* concatenate a rotation by angle around the Y axis to the transformation */
/* in xformtochange, placing result back in xformtochange                  */
void AppendRotationY(double (*XformToChange)[4],double Angle)
{
    double Temp00,Temp01,Temp02,Temp20,Temp21,Temp22;
    double CosTemp = cos(Angle), SinTemp = sin(Angle);
    /* Calculate the new values of the four affected matrix entries */
    Temp00 = CosTemp*XformToChange[0][0] + SinTemp*XformToChange[2][0];
    Temp01 = CosTemp*XformToChange[0][1] + SinTemp*XformToChange[2][1];
    Temp02 = CosTemp*XformToChange[0][2] + SinTemp*XformToChange[2][2];
    Temp20 = -SinTemp*XformToChange[0][0] + CosTemp*XformToChange[2][0];
    Temp21 = -SinTemp*XformToChange[0][1] + CosTemp*XformToChange[2][1];
    Temp22 = -SinTemp*XformToChange[0][2] + CosTemp*XformToChange[2][2];
    /* Put the results back into XformToChange */
    XformToChange[0][0] = Temp00;
    XformToChange[0][1] = Temp01;
    XformToChange[0][2] = Temp02;
    XformToChange[2][0] = Temp20;
    XformToChange[2][1] = Temp21;
    XformToChange[2][2] = Temp22;
          
}
