#include <stdio.h>
#include <stdlib.h>
#include <SDL/SDL.h>
#include <math.h>
#include "types.h"
#include "polygon.h"
#include "vl.h"

#define ROTATION (M_PI / 30.0)    /* rotate by 6 degrees at a time */

/* transformation from cube object space to world space. Initially  */
/* set up to perform no rotation and to move the cube into world    */
/* space -100 units away from the origin down the z axis. Given the */
/* viewing point, -100 down the z axis means 100 units away in the  */
/* direction of view. The program dynamically changes both the      */
/* translation and the rotation                                     */
static double CubeWorldXform[4][4] =
{
    {1.0,0.0,0.0,0.0},
    {0.0,1.0,0.0,0.0},
    {0.0,0.0,1.0,-100.0},
    {0.0,0.0,0.0,1.0}       
};
/* Transformation from world space into view space. Because in this    */
/* application the view point is fixed at the origin of world space,   */
/* looking down the z axis in the direction of increasing Z, view      */
/* space is identical to world space, and this is the identity matrix  */
static double WorldViewXform[4][4] =
{
    {1.0,0.0,0.0,0.0},
    {0.0,1.0,0.0,0.0},
    {0.0,0.0,1.0,0.0},
    {0.0,0.0,0.0,1.0}        
};

/* all vertices in the cube */
static struct Point3 CubeVerts[] = 
{
    {15,15,15,1},{15,15,-15,1},{15,-15,15,1},{15,-15,-15,1},
    {-15,15,15,1},{-15,15,-15,1},{-15,-15,15,1},{-15,-15,-15,1}         
};
/* vertices after transformation */
static struct Point3 XformedCubeVerts[sizeof(CubeVerts) / sizeof(struct Point3)];

/* vertices after projection */
static struct Point3 ProjectedCubeVerts[sizeof(CubeVerts) / sizeof(struct Point3)];

/* vertices in screen coordinates */
static struct Point ScreenCubeVerts[sizeof(CubeVerts) / sizeof(struct Point3)];

/* vertex indices for individual faces */
static S32 Face1[] = {1,3,2,0};
static S32 Face2[] = {5,7,3,1};
static S32 Face3[] = {4,5,1,0};
static S32 Face4[] = {3,7,6,2};
static S32 Face5[] = {5,4,6,7};
static S32 Face6[] = {0,2,6,4};

/* list of cube faces */
static struct Face CubeFaces[] = 
{
    {Face1,4,1},{Face2,4,2},{Face3,4,3},
    {Face4,4,4},{Face5,4,5},{Face6,4,6}     
};

/* master description for cube */
static struct Object Cube =
{
    sizeof(CubeVerts) / sizeof(struct Point3),
    CubeVerts,
    XformedCubeVerts,
    ProjectedCubeVerts,
    ScreenCubeVerts,
    sizeof(CubeFaces) / sizeof(struct Face),
    CubeFaces       
};

int main(int argc, char *argv[])
{
  S32 RecalcXform = 1;  
  double WorkingXform[4][4];
  SDL_Event event = {0};
  input_type UserControls;
  memset(&UserControls,0,sizeof(UserControls));
  
  /* try and get a handle to the screen */
  if(InitSDLScreenBuf() == 1)
  {
      /* cant get handle to screen so exit */
      return(1);
  }
    
  /* keep rotating the polygon */
  while(UserControls.Exit == 0)
  {
      /* clear last user input */
      memset(&UserControls,0,sizeof(UserControls));
      /* then see if user was pressed a key */
      GetUserInput(&UserControls);
         
      /* left (-x) */ 
      if(UserControls.Left == 1)
      {
          CubeWorldXform[0][3] -= 3.0;
          RecalcXform = 1;                         
      }
      /* right (+x) */
      if(UserControls.Right == 1)
      {
          CubeWorldXform[0][3] += 3.0;
          RecalcXform = 1;                         
      }
      /* up (+y) */
      if(UserControls.Forward == 1)
      {
          CubeWorldXform[1][3] += 3.0;
          RecalcXform = 1;                         
      }
      /* down (-y) */
      if(UserControls.Back == 1)
      {
          CubeWorldXform[1][3] -= 3.0;
          RecalcXform = 1;                         
      }
      
      AppendRotationY(CubeWorldXform,ROTATION);
      RecalcXform = 1;
      
      /* regenerate the object->view transformation and */
      /* retransform / project if necessary             */
      if(RecalcXform == 1)
      {
           ConcatXforms(WorldViewXform, CubeWorldXform, WorkingXform);
           XformAndProjectPoints(WorkingXform, &Cube);                   
      }
      
      /* Draw all visible faces of the cube */
      DrawVisibleFaces(&Cube);
      
      /* update screen */
      Redraw();
            
      /* clear screen buffer for next write */
      Cls();

      SDL_Delay(50); /* so it consumes less processing power */                                                   
  }
}
